from dataclasses import dataclass, field
from typing import List, Optional
from sgrspecification.generic.base_types import (
    AlternativeNames,
    DataDirectionFunctionalProfile,
    DataTypeFunctionalProfile,
    FunctionalProfileIdentification,
    GenericAttributeListFunctionalProfile,
    LegibleDescription,
    PresenceLevel,
    ReleaseNotes,
    RequestParam,
    Units,
)

__NAMESPACE__ = "http://www.smartgridready.com/ns/V0/"


@dataclass
class FunctionalProfileDataPoint:
    """
    Data point element.
    """
    data_point: Optional["FunctionalProfileDataPoint.DataPoint"] = field(
        default=None,
        metadata={
            "name": "dataPoint",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    generic_attribute_list: Optional[GenericAttributeListFunctionalProfile] = field(
        default=None,
        metadata={
            "name": "genericAttributeList",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )

    @dataclass
    class DataPoint:
        """
        Generic data point description.

        :ivar data_point_name: Bezeichnung des Datenpunktes:
            «Schlagwort», welches die Bedeutung identifiziert. Diese
            Bezeichnung gilt neben der Indexnummer als Definition für
            den SmartGridready Namespace für die maschinenlesbaren
            Daten.
        :ivar data_direction:
        :ivar presence_level:
        :ivar request_params:
        :ivar data_type:
        :ivar unit:
        :ivar array_length:
        :ivar alternative_names:
        :ivar legible_description: Published and printable information
            related to this datapoint
        """
        data_point_name: Optional[str] = field(
            default=None,
            metadata={
                "name": "dataPointName",
                "type": "Element",
                "namespace": "http://www.smartgridready.com/ns/V0/",
                "required": True,
            }
        )
        data_direction: Optional[DataDirectionFunctionalProfile] = field(
            default=None,
            metadata={
                "name": "dataDirection",
                "type": "Element",
                "namespace": "http://www.smartgridready.com/ns/V0/",
                "required": True,
            }
        )
        presence_level: Optional[PresenceLevel] = field(
            default=None,
            metadata={
                "name": "presenceLevel",
                "type": "Element",
                "namespace": "http://www.smartgridready.com/ns/V0/",
                "required": True,
            }
        )
        request_params: Optional[RequestParam] = field(
            default=None,
            metadata={
                "name": "requestParams",
                "type": "Element",
                "namespace": "http://www.smartgridready.com/ns/V0/",
            }
        )
        data_type: Optional[DataTypeFunctionalProfile] = field(
            default=None,
            metadata={
                "name": "dataType",
                "type": "Element",
                "namespace": "http://www.smartgridready.com/ns/V0/",
                "required": True,
            }
        )
        unit: Optional[Units] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "http://www.smartgridready.com/ns/V0/",
                "required": True,
            }
        )
        array_length: Optional[int] = field(
            default=None,
            metadata={
                "name": "arrayLength",
                "type": "Element",
                "namespace": "http://www.smartgridready.com/ns/V0/",
            }
        )
        alternative_names: Optional[AlternativeNames] = field(
            default=None,
            metadata={
                "name": "alternativeNames",
                "type": "Element",
                "namespace": "http://www.smartgridready.com/ns/V0/",
            }
        )
        legible_description: List[LegibleDescription] = field(
            default_factory=list,
            metadata={
                "name": "legibleDescription",
                "type": "Element",
                "namespace": "http://www.smartgridready.com/ns/V0/",
                "max_occurs": 4,
            }
        )


@dataclass
class FunctionalProfileDataPointList:
    """
    List of data points.
    """
    data_point_list_element: List[FunctionalProfileDataPoint] = field(
        default_factory=list,
        metadata={
            "name": "dataPointListElement",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class FunctionalProfileFrame:
    """
    Functional profile template.
    """
    class Meta:
        namespace = "http://www.smartgridready.com/ns/V0/"

    release_notes: Optional[ReleaseNotes] = field(
        default=None,
        metadata={
            "name": "releaseNotes",
            "type": "Element",
        }
    )
    functional_profile: Optional["FunctionalProfileFrame.FunctionalProfile"] = field(
        default=None,
        metadata={
            "name": "functionalProfile",
            "type": "Element",
            "required": True,
        }
    )
    generic_attribute_list: Optional[GenericAttributeListFunctionalProfile] = field(
        default=None,
        metadata={
            "name": "genericAttributeList",
            "type": "Element",
        }
    )
    data_point_list: Optional[FunctionalProfileDataPointList] = field(
        default=None,
        metadata={
            "name": "dataPointList",
            "type": "Element",
        }
    )

    @dataclass
    class FunctionalProfile:
        """
        Functional profile element.

        :ivar functional_profile_identification:
        :ivar alternative_names:
        :ivar legible_description: Published and printable information
            related to this functional profile
        """
        functional_profile_identification: Optional[FunctionalProfileIdentification] = field(
            default=None,
            metadata={
                "name": "functionalProfileIdentification",
                "type": "Element",
                "required": True,
            }
        )
        alternative_names: Optional[AlternativeNames] = field(
            default=None,
            metadata={
                "name": "alternativeNames",
                "type": "Element",
            }
        )
        legible_description: List[LegibleDescription] = field(
            default_factory=list,
            metadata={
                "name": "legibleDescription",
                "type": "Element",
                "max_occurs": 4,
            }
        )



from dataclasses import dataclass, field
from typing import List, Optional
from sgrspecification.generic.base_types import (
    AlternativeNames,
    DataDirectionProduct,
    DataTypeProduct,
    GenericAttributeListProduct,
    LegibleDescription,
    Units,
)

__NAMESPACE__ = "http://www.smartgridready.com/ns/V0/"


@dataclass
class DataPointDescription:
    """
    Generic data point properties.

    :ivar data_point_name: Name of the data point (unique on functional
        profile)
    :ivar data_direction:
    :ivar data_type:
    :ivar value:
    :ivar unit:
    :ivar array_length: Optional, if present the data point is an array
        of specified length
    :ivar minimum_value:
    :ivar maximum_value:
    :ivar unit_conversion_multiplicator:
    :ivar alternative_names:
    :ivar legible_description: Published and printable information
        related to this data point
    :ivar programmer_hints: additional device-specific implementation
        hints for this data point
    """
    data_point_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "dataPointName",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    data_direction: Optional[DataDirectionProduct] = field(
        default=None,
        metadata={
            "name": "dataDirection",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    data_type: Optional[DataTypeProduct] = field(
        default=None,
        metadata={
            "name": "dataType",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    value: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    unit: Optional[Units] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    array_length: Optional[int] = field(
        default=None,
        metadata={
            "name": "arrayLength",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    minimum_value: Optional[float] = field(
        default=None,
        metadata={
            "name": "minimumValue",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    maximum_value: Optional[float] = field(
        default=None,
        metadata={
            "name": "maximumValue",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    unit_conversion_multiplicator: Optional[float] = field(
        default=None,
        metadata={
            "name": "unitConversionMultiplicator",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    alternative_names: Optional[AlternativeNames] = field(
        default=None,
        metadata={
            "name": "alternativeNames",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    legible_description: List[LegibleDescription] = field(
        default_factory=list,
        metadata={
            "name": "legibleDescription",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "max_occurs": 4,
        }
    )
    programmer_hints: List[LegibleDescription] = field(
        default_factory=list,
        metadata={
            "name": "programmerHints",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "max_occurs": 4,
        }
    )


@dataclass
class DataPointBase:
    """
    Data point element.
    """
    data_point: Optional[DataPointDescription] = field(
        default=None,
        metadata={
            "name": "dataPoint",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    generic_attribute_list: Optional[GenericAttributeListProduct] = field(
        default=None,
        metadata={
            "name": "genericAttributeList",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )



from dataclasses import dataclass, field
from typing import List, Optional
from sgrspecification.generic.base_type_level_of_operation_type import LevelOfOperation
from sgrspecification.generic.base_types import (
    AlternativeNames,
    DataTypeProduct,
    DeviceCategory,
    GenericAttributeListProduct,
    LegibleDescription,
    PowerSource,
    ReleaseNotes,
    VersionNumber,
)
from sgrspecification.product.contact_interface import ContactInterface
from sgrspecification.product.generic_interface import GenericInterface
from sgrspecification.product.modbus_interface import ModbusInterface
from sgrspecification.product.rest_api_interface import RestApiInterface

__NAMESPACE__ = "http://www.smartgridready.com/ns/V0/"


@dataclass
class ConfigurationDescription(LegibleDescription):
    label: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class DeviceInformation:
    """
    :ivar alternative_names:
    :ivar legible_description: Published and printable information
        related to this product
    :ivar device_category:
    :ivar is_local_control: Value "false" means "is cloud control
        device", indicating that this service is based on cloud. "True"
        indicates that services are provided within the range of the
        local area.
    :ivar software_revision: software version information for this
        product declaration
    :ivar hardware_revision: hardware version information for this
        product declaration
    :ivar brand_name: branding information
    :ivar power_source: power supply type
    :ivar nominal_power: nominal Power of the device (installation)
    :ivar manufacturer_specification_identification: manufacturers
        specification identifier
    :ivar manufacturer_label: manufacturers label of the device
    :ivar general_remarks: author of this sheet may add remarks / non
        disclaimer statements
    :ivar level_of_operation: defines the SGr Label Level 1...6 of the
        highest level functional profile of this device
    :ivar version_number:
    :ivar programmer_hints: additional device-specific implementation
        hints for this device
    """
    alternative_names: Optional[AlternativeNames] = field(
        default=None,
        metadata={
            "name": "alternativeNames",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    legible_description: List[LegibleDescription] = field(
        default_factory=list,
        metadata={
            "name": "legibleDescription",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "max_occurs": 4,
        }
    )
    device_category: Optional[DeviceCategory] = field(
        default=None,
        metadata={
            "name": "deviceCategory",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    is_local_control: Optional[bool] = field(
        default=None,
        metadata={
            "name": "isLocalControl",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    software_revision: Optional[str] = field(
        default=None,
        metadata={
            "name": "softwareRevision",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    hardware_revision: Optional[str] = field(
        default=None,
        metadata={
            "name": "hardwareRevision",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    brand_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "brandName",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    power_source: Optional[PowerSource] = field(
        default=None,
        metadata={
            "name": "powerSource",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    nominal_power: Optional[str] = field(
        default=None,
        metadata={
            "name": "nominalPower",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    manufacturer_specification_identification: Optional[str] = field(
        default=None,
        metadata={
            "name": "manufacturerSpecificationIdentification",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    manufacturer_label: Optional[str] = field(
        default=None,
        metadata={
            "name": "manufacturerLabel",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    general_remarks: Optional[str] = field(
        default=None,
        metadata={
            "name": "generalRemarks",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    level_of_operation: Optional[LevelOfOperation] = field(
        default=None,
        metadata={
            "name": "levelOfOperation",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    version_number: Optional[VersionNumber] = field(
        default=None,
        metadata={
            "name": "versionNumber",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    programmer_hints: List[LegibleDescription] = field(
        default_factory=list,
        metadata={
            "name": "programmerHints",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "max_occurs": 4,
        }
    )


@dataclass
class InterfaceList:
    """
    List of supported interfaces.
    """
    modbus_interface: Optional[ModbusInterface] = field(
        default=None,
        metadata={
            "name": "modbusInterface",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    rest_api_interface: Optional[RestApiInterface] = field(
        default=None,
        metadata={
            "name": "restApiInterface",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    contact_interface: Optional[ContactInterface] = field(
        default=None,
        metadata={
            "name": "contactInterface",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    generic_interface: Optional[GenericInterface] = field(
        default=None,
        metadata={
            "name": "genericInterface",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class ConfigurationListElement:
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    data_type: Optional[DataTypeProduct] = field(
        default=None,
        metadata={
            "name": "dataType",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    configuration_description: List[ConfigurationDescription] = field(
        default_factory=list,
        metadata={
            "name": "configurationDescription",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "max_occurs": 4,
        }
    )


@dataclass
class ConfigurationList:
    configuration_list_element: List[ConfigurationListElement] = field(
        default_factory=list,
        metadata={
            "name": "configurationListElement",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class DeviceFrame:
    """
    Product declaration.

    :ivar device_name: Device Name in the context of the manufacturer
    :ivar manufacturer_name: Name of the Manufacturer or OEM
    :ivar specification_owner_identification:
    :ivar release_notes:
    :ivar device_information:
    :ivar configuration_list:
    :ivar generic_attribute_list:
    :ivar interface_list:
    """
    class Meta:
        namespace = "http://www.smartgridready.com/ns/V0/"

    device_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "deviceName",
            "type": "Element",
            "required": True,
        }
    )
    manufacturer_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "manufacturerName",
            "type": "Element",
        }
    )
    specification_owner_identification: Optional[str] = field(
        default=None,
        metadata={
            "name": "specificationOwnerIdentification",
            "type": "Element",
            "required": True,
        }
    )
    release_notes: Optional[ReleaseNotes] = field(
        default=None,
        metadata={
            "name": "releaseNotes",
            "type": "Element",
            "required": True,
        }
    )
    device_information: Optional[DeviceInformation] = field(
        default=None,
        metadata={
            "name": "deviceInformation",
            "type": "Element",
            "required": True,
        }
    )
    configuration_list: Optional[ConfigurationList] = field(
        default=None,
        metadata={
            "name": "configurationList",
            "type": "Element",
        }
    )
    generic_attribute_list: Optional[GenericAttributeListProduct] = field(
        default=None,
        metadata={
            "name": "genericAttributeList",
            "type": "Element",
        }
    )
    interface_list: Optional[InterfaceList] = field(
        default=None,
        metadata={
            "name": "interfaceList",
            "type": "Element",
            "required": True,
        }
    )



from pymodbus.constants import Endian
from xsdata.formats.dataclass.context import XmlContext
from xsdata.formats.dataclass.parsers import XmlParser

from sgrspecification.product import DeviceFrame, BitOrder, ModbusDataPoint
from sgr_library.exceptions import DataPointException, FunctionalProfileException


# TODO error handling
def get_protocol(xml_file: str) -> str:
    """
    Searches for protocol type in xml file
    :return: protocol type string
    """
    parser = XmlParser(context=XmlContext())
    root = parser.parse(xml_file, DeviceFrame)

    restapi = root.interface_list.rest_api_interface
    generic = root.interface_list.generic_interface
    contact = root.interface_list.contact_interface
    modbus = root.interface_list.modbus_interface
    if modbus:
        return "modbus"
    elif restapi:
        return "restapi"
    elif generic:
        return "generic"
    elif contact:
        return "contact"
    else:
        return "error"


# TODO make this one so that it is generic.
def find_dp(root, fp_name: str, dp_name: str) -> ModbusDataPoint:
    """
    Searches the datapoint in the root element.
    :param root: The root element created with the xsdata parser
    :param fp_name: The name of the funcitonal profile in which the datapoint resides
    :param dp_name: The name of the datapoint
    :returns: The datapoint element found in root, if not, returns None.
    """
    for fp in root.interface_list.modbus_interface.functional_profile_list.functional_profile_list_element:
        for dp in fp.data_point_list.data_point_list_element:
            if fp.functional_profile.functional_profile_name == fp_name and dp.data_point.data_point_name:
                return dp
    raise DataPointException(f"Datapoint {dp_name} not found in functional profile {fp_name}.")
    # raise FunctionalProfileException(f"Functional profile {fp_name} not found in XML file.")


def get_modbusInterfaceSelection(xml_file: str) -> str:
    try:
        parser = XmlParser(context=XmlContext())
        root = parser.parse(xml_file, DeviceFrame)
        interface_selection = root.interface_list.modbus_interface.modbus_interface_description.modbus_interface_selection.value
        if interface_selection not in ['RTU', 'TCPIP', 'UDPIP', 'RTU-ASCII', 'TCPIP-ASCII', 'UDPIP-ASCII']:
            raise ValueError('Invalid Modbus interface selection found.')
        return interface_selection
    except Exception as e:
        raise ValueError(f"Error parsing XML file or extracting Modbus interface selection: {e}")


def get_address(root) -> str:
    try:
        return str(root.interface_list.modbus_interface.modbus_interface_description.modbus_tcp.address)
    except AttributeError:
        raise ValueError("IP address not found in XML file.")


def get_port(root) -> int:
    try:
        return int(root.interface_list.modbus_interface.modbus_interface_description.modbus_tcp.port)
    except (AttributeError, ValueError):
        raise ValueError("Port not found or invalid in XML file.")


def get_slave(root) -> int:
    try:
        return int(root.interface_list.modbus_interface.modbus_interface_description.modbus_tcp.slave_id)
    except (AttributeError, ValueError):
        raise ValueError("Slave ID not found or invalid in XML file.")


def get_slave_rtu(root) -> int:
    try:
        return int(root.interface_list.modbus_interface.modbus_interface_description.modbus_rtu.slave_addr)
    except (AttributeError, ValueError):
        raise ValueError("RTU Slave ID not found or invalid in XML file.")


def get_endian(root) -> str:
    try:
        match root.interface_list.modbus_interface.modbus_interface_description.bit_order:
            case BitOrder.BIG_ENDIAN:
                return Endian.BIG
            case _:
                return Endian.LITTLE
    except AttributeError:
        raise ValueError("Endian type not found in XML file.")


def get_baudrate(root) -> int:
    try:
        return int(
            root.interface_list.modbus_interface.modbus_interface_description.modbus_rtu.baud_rate_selected.value)
    except (AttributeError, ValueError):
        raise ValueError("Baudrate not found or invalid in XML file.")


def get_parity(root) -> str:
    try:
        parity_string = str(
            root.interface_list.modbus_interface.modbus_interface_description.modbus_rtu.parity_selected.value)
        if parity_string == "EVEN":
            return "E"
        else:
            raise NotImplementedError(f'Parity type "{parity_string}" not supported.')
    except AttributeError:
        raise ValueError("Parity type not found in XML file.")



import logging
import os
from dataclasses import dataclass
from typing import Optional, Tuple, Dict, Any, Iterable
from pymodbus.constants import Endian
from xsdata.formats.dataclass.parsers import XmlParser
from xsdata.formats.dataclass.context import XmlContext
import time

from sgr_library.api import BaseSGrInterface, DeviceInformation, FunctionProfile, DataPointProtocol, DataPoint, \
    build_configurations_parameters, ConfigurationParameter
from sgr_library.auxiliary_functions import find_dp
from sgr_library.converters import build_converter
from sgrspecification.generic import Parity, DataDirectionProduct

# from sgr_library.data_classes.ei_modbus import SgrModbusDeviceDescriptionType
from sgrspecification.product import DeviceFrame, ModbusDataPoint, ModbusFunctionalProfile

from sgr_library.modbusRTU_client_async import SGrModbusRTUClient
from sgr_library.validators import build_validator


def get_port(root) -> str:
    """
    :param root: The root element created with the xsdata parser
    :returns: string with port from xml file.
    """
    return (str(root.modbus_interface_desc.trsp_srv_modbus_tcpout_of_box.port))
    # return(str(root.modbus_interface_desc.trspSrvModbusRTUoutOfBox.port)) #TODO Port datapoint for RTU in XML is not existing


def get_slave(root) -> int:
    """
    returns the selected slave address
    """
    return root.interface_list.modbus_interface.modbus_interface_description.modbus_rtu.slave_addr


def get_endian(root) -> Endian:
    return Endian.BIG


def get_baudrate(root) -> int:
    """
    returns the selected baudrate. Implemented for Even
    """
    return root.interface_list.modbus_interface.modbus_interface_description.modbus_rtu.baud_rate_selected.value


def get_parity(root) -> str:
    """
    returns the parity. Implemented for Even
    """
    parity = root.interface_list.modbus_interface.modbus_interface_description.modbus_rtu.parity_selected
    match parity:
        case Parity.EVEN:
            return 'E'
        case _:
            raise NotImplementedError


def build_modbus_rtu_data_point(data_point: ModbusDataPoint, function_profile: ModbusFunctionalProfile,
                                interface: 'SgrModbusRtuInterface') -> DataPoint:
    protocol = ModBusRTUDataPoint(data_point, function_profile, interface)
    converter = build_converter(data_point.data_point.unit)
    validator = build_validator(data_point.data_point.data_type)
    return DataPoint(protocol, converter, validator)


class ModBusRTUDataPoint(DataPointProtocol):

    def __init__(self, modbus_api_dp: ModbusDataPoint, modbus_api_fp: ModbusFunctionalProfile,
                 interface: 'SgrModbusRtuInterface'):
        self._dp = modbus_api_dp
        self._fp = modbus_api_fp
        self._interface = interface

    async def write(self, data: Any):
        return await self._interface.setval(self.name()[0], self.name()[1], data)

    async def read(self) -> Any:
        return await self._interface.getval(self.name()[0], self.name()[1])

    def name(self) -> tuple[str, str]:
        return self._fp.functional_profile.functional_profile_name, self._dp.data_point.data_point_name

    def direction(self) -> DataDirectionProduct:
        return self._dp.data_point.data_direction


class ModBusRTUFunctionProfile(FunctionProfile):

    def __init__(self, modbus_api_fp: ModbusFunctionalProfile, interface: 'SgrModbusRtuInterface'):
        self._fp = modbus_api_fp
        self._interface = interface
        dps = [build_modbus_rtu_data_point(dp, self._fp, self._interface) for dp in
               self._fp.data_point_list.data_point_list_element]
        self._data_points = {dp.name(): dp for dp in dps}

    def name(self) -> str:
        return self._fp.functional_profile.functional_profile_name

    def get_data_points(self) -> dict[tuple[str, str], DataPoint]:
        return self._data_points


class SgrModbusRtuInterface(BaseSGrInterface):
    # a global Modbus client
    globalModbusRTUClient = None

    def __init__(self, frame: DeviceFrame) -> None:
        """
        Creates a connection from xml file data.
        Parses the xml file with xsdata library.
        :param xml_file: Name of the xml file to parse
        """
        self.root = frame
        # self.root = parser.parse(interface_file, SgrModbusDeviceDescriptionType)
        self.port = os.getenv("SGR_RTU_PORT")  # get_port(self.root) #TODO überlegungen machen wo Port untergebracht wird
        self._configurations_params = build_configurations_parameters(frame.configuration_list)
        self.baudrate = get_baudrate(self.root)
        self.parity = get_parity(self.root)
        self.slave_id = get_slave(self.root)
        self.byte_order = get_endian(self.root)
        fps = [ModBusRTUFunctionProfile(profile, self) for profile in
               self.root.interface_list.modbus_interface.functional_profile_list.functional_profile_list_element]
        self._function_profiles = {fp.name(): fp for fp in fps}
        self._device_information = DeviceInformation(
            name=frame.device_name,
            manufacture=frame.manufacturer_name,
            software_revision=frame.device_information.software_revision,
            hardware_revision=frame.device_information.hardware_revision,
            device_category=frame.device_information.device_category,
            is_local=frame.device_information.is_local_control
        )

    def get_pymodbus_client(self):
        """
        returns the pymodbus client object
        """
        return self.client.client

    # TODO
    def get_dp_attribute(self, datapoint: str, attribute: str):
        """
        Searches for a specific attribute in the datapoint via a key.
        :param attribute"address", "size", "bitrank", "data_type", "register_type", "unit", "multiplicator", "power_of", "name"
        :returns: The chosen attribute.
        """
        # TODO
        ...

    # TODO assign multiple dispatch to the function.
    '''def getval(self, fp_name: str, dp_name: str) -> float:
        """
        Reads datapoint value.
        :param fp_name: The name of the funcitonal profile in which the datapoint resides.
        :param dp_name: The name of the datapoint.
        :returns: The current decoded value in the datapoint register.
        """
        dp = find_dp(self.root, fp_name, dp_name)
        address = self.get_address(dp)
        size = self.get_size(dp)
        data_type = self.get_datatype(dp)
        reg_type = self.get_register_type(dp)
        return self.client.value_decoder(address, size, data_type, reg_type)'''

    # getval with multiple dispatching
    async def getval(self, *parameter) -> float:
        """
        Reads datapoint value.

        :dp: The already obtained datapoint object

        2 parameters alternative:
        :param fp_name: The name of the functional profile in which the datapoint resides.
        :param dp_name: The name of the datapoint.

        :returns: The current decoded value in the datapoint register.
        """
        if len(parameter) == 2:
            dp = find_dp(self.root, parameter[0], parameter[1])
        else:
            dp = parameter[0]
        address = self.get_address(dp)
        size = self.get_size(dp)
        data_type = self.get_datatype(dp)
        reg_type = self.get_register_type(dp)
        slave_id = self.slave_id
        order = self.byte_order

        logging.debug(
            f"getVal() with address: {address}, size: {size}, data_type:{data_type}, slave_id: {slave_id}, order: {order}")  # for debugging
        return await self.client.value_decoder(address, size, data_type, reg_type, slave_id, order)

    async def setval(self, fp_name: str, dp_name: str, value: float) -> None:
        """
        Writes datapoint value.
        :param fp_name: The name of the funcitonal profile in which the datapoint resides.
        :param dp_name: The name of the datapoint.
        :param value: The value that is to be written on the datapoint.
        """
        dp: ModbusDataPoint = find_dp(self.root, fp_name, dp_name)
        address = self.get_address(dp)
        data_type = self.get_datatype(dp)
        slave_id = self.slave_id
        order = self.byte_order
        await self.client.value_encoder(address, value, data_type, slave_id, order)

    def get_device_profile(self):
        return (self.root.device_profile)

    def get_register_type(self, dp: ModbusDataPoint) -> str:
        """
        Returns register type E.g. "HoldRegister"
        :param fp_name: The name of the functional profile
        :param dp_name: The name of the data point.
        :returns: The type of the register
        """
        return dp.modbus_data_point_configuration.register_type.value.__str__()

    def get_datatype(self, dp: ModbusDataPoint) -> str:
        datatype = dp.modbus_data_point_configuration.modbus_data_type.__dict__
        for key in datatype:
            if datatype[key] is not None:
                return key

    def get_bit_rank(self, dp: ModbusDataPoint):
        return dp.modbus_data_point_configuration.bit_rank

    def get_address(self, dp: ModbusDataPoint):
        return dp.modbus_data_point_configuration.address

    def get_size(self, dp: ModbusDataPoint):
        return dp.modbus_data_point_configuration.number_of_registers

    def get_multiplicator(self, dp: ModbusDataPoint) -> int:
        return dp.modbus_attributes.scaling_factor.multiplicator

    def get_power_10(self, dp: ModbusDataPoint) -> int:
        return dp.modbus_attributes.scaling_factor.powerof10

    def get_unit(self, dp: ModbusDataPoint):
        return dp.data_point.unit.value

    def get_name(self, dp: ModbusDataPoint):
        return dp.data_point.data_point_name

    """
    def find_dp(self, fp_name: str, dp_name: str) -> SgrModbusDataPointType:
        
        Searches the datapoint in the root element.
        :param root: The root element created with the xsdata parser
        :param fp_name: The name of the funcitonal profile in which the datapoint resides
        :param dp_name: The name of the datapoint
        :returns: The datapoint element found in root, if not, returns None.
        
        for fp in self.root.fp_list_element:
                if fp_name == fp.functional_profile.profile_name:
                    #Secondly we filter the datpoint name
                    for dp in fp.dp_list_element:
                        if dp_name == dp.data_point[0].datapoint_name:
                            return dp
        return None
    """

    def configuration_parameter(self) -> list[ConfigurationParameter]:
        return self._configurations_params

    async def connect(self):
        # check if there already exists a ModbusRTUClient for the communication over ModbusRTU
        if SgrModbusRtuInterface.globalModbusRTUClient is None:
            self.client = SGrModbusRTUClient(str(self.port), str(self.parity), int(self.baudrate))
            SgrModbusRtuInterface.globalModbusRTUClient = self.client
        else:
            self.client = SgrModbusRtuInterface.globalModbusRTUClient

        if self.client:
            connected = await self.client.connect()
            if connected:
                print("Connected to ModbusRTU on Port: " + self.port)

    def get_function_profiles(self) -> dict[str, FunctionProfile]:
        return self._function_profiles

    def device_information(self) -> DeviceInformation:
        return self._device_information

    def find_dp(self, fp_name: str, dp_name: str):
        """
        Searches the datapoint in the root element.
        :param root: The root element created with the xsdata parser
        :param fp_name: The name of the funcitonal profile in which the datapoint resides
        :param dp_name: The name of the datapoint
        :returns: The datapoint element found in root, if not, returns None.
        """
        fp = next(filter(lambda x: x.functional_profile.profile_name == fp_name, self.root.fp_list_element), None)
        if fp:
            dp = next(filter(lambda x: x.data_point.datapoint_name == dp_name, fp.dp_list_element), None)
            if dp:
                return dp

    def get_device_name(self) -> str:
        """
        returns the s_LV1_Name of the device
        """
        return self.root.device_profile.dev_name_list.s_lv1_name

    def get_modbusInterfaceSelection(self) -> str:
        """
        Returns selected Modbus Interface in XML file
        """
        return (self.root.modbus_interface_desc.modbus_interface_selection.value)

    def get_manufacturer(self):
        """
        returns the manufacturer name of the device
        """
        # TODO it could also be root.device_information.alternative_names.manuf_name
        return self.root.manufacturer_name

    def set_slave_id(self, slave_id: int):
        """
        changes the slave id for the instance
        """
        self.slave_id = slave_id

    # TODO a getval for L1, L2 and L3 at the same time


if __name__ == "__main__":
    starting_time = time.time()
    print('start')
    interface_file = '../xml_files/SGr_04_0016_xxxx_ABBMeterV0.2.1.xml'
    a = SgrModbusRtuInterface(interface_file)
    print('ActivePowerACtot :', a.getval('ActivePowerAC', 'ActivePowerACtot'))
    # Power = a.client.value_decoder(0x5B14,2,"int32","HoldingRegister",1,Endian.Big)
    # print(str(Power*0.01) + "W")

    # a.setval('ActiveEnerBalanceAC', 'ActiveImportAC', 9000)

    dp = find_dp(a.root, 'ActiveEnerBalanceAC', 'ActiveImportAC')
    print("ActiveImportAC :", a.getval(dp))

    a.client.client.close()
    print("print finished")



from enum import Enum

__NAMESPACE__ = "http://www.smartgridready.com/ns/V0/"


class FunctionalProfileCategory(Enum):
    """
    :cvar BATTERY: 1: Batterie
    :cvar COMPRESSOR: 2: Kompressor
    :cvar DEVICE_INFORMATION: 3: Geräteinformation
    :cvar DHWCIRCUIT: 4: WarmWasserHeizung
    :cvar DHWSTORAGE: 5: WarmWasserSpeicher
    :cvar DISHWASHER: 6: Geschirrspülmaschine
    :cvar DRYER: 7: Trockner
    :cvar ELECTRICAL_IMMERSION_HEATER: 8: elTauchSieder
    :cvar FAN: 9: Ventilator
    :cvar GAS_HEATING_APPLIANCE: 10: GasWärmeAnwendung
    :cvar METERING: 11: allgemein verwendbar
    :cvar HEATING_BUFFER_STORAGE: 12: WaermeBufferSpeicher
    :cvar HEATING_CIRCUIT: 13: Heizkreis
    :cvar HEATING_OBJECT: 14: Heizungsobjekt
    :cvar HEATING_ZONE: 15: Heizzone
    :cvar HEAT_PUMP_CONTROL: 16: WärmePumpenAnwendung
    :cvar HEAT_SINK_CIRCUIT: 17: KuehlKoerperSchaltung
    :cvar HEAT_SOURCE_CIRCUIT: 18: WärmekörperSchaltung
    :cvar HEAT_SOURCE_UNIT: 19: Wärmequelle
    :cvar HVACCONTROLLER: 20: HLKController
    :cvar HVACROOM: 21: HLKRaum
    :cvar INSTANT_DHWHEATER: 22: DurchflussWarmwasserHeizung
    :cvar INVERTER: 23: Wechselrichter / Wandler
    :cvar OIL_HEATING_APPLIANCE: 24: OelWaermeAnwendung
    :cvar PUMP: 25: Pumpe
    :cvar REFRIGERANT_CIRCUIT: 26: Kältemittelkreislauf
    :cvar SMART_ENERGY_APPLIANCE: 27: SmartEnergyAppliance
    :cvar SOLAR_DHWSTORAGE: 28: SolarWarmwasserSpeicher
    :cvar SOLAR_THERMAL_CIRCUIT: SolarWarmwasserKreislauf
    :cvar SUB_METER_ELECTRICITY: 30: SubMeter Elektrisch
    :cvar TEMPERATURE_SENSOR: 31: Temperatursensor
    :cvar WASHER: 32: Waschmaschine
    :cvar BATTERY_SYSTEM: 33: Batteriesystem
    :cvar ELECTRICITY_GENERATION_SYSTEM: 34: Generator Elektrisch
    :cvar ELECTRICITY_STORAGE_SYSTEM: 35: ElektroSpeicher System
    :cvar SGCP: 36: GridConnectionPoint Nachbarschaft
    :cvar HOUSEHOLD: 37: Haushalt
    :cvar PVSYSTEM: 38: PV System
    :cvar EV: 39: Elektromobil
    :cvar EVSE: 40: Ladestationscontroller für Elektrofahrzeug
    :cvar CHARGING_OUTLET: 41: Ladestation
    :cvar CEM: 42: Energiemanager für Endverbraucher
    :cvar ACTUATOR: 43:allgemein fuer Aktor
    :cvar SENSOR: 44: allgemein fuer Messdatenaufnahme
    :cvar CONTROLLER: 45:Steuergeraet
    :cvar ENV_CONDITION: 46:Umweltbedingungen
    :cvar FLEX_BUILDING_CAMPUS: 47: Gebäude und Campus.elektrische
        Flexibilität
    :cvar DYNAMIC_TARIFF: 49: Dynamic Tariff
    :cvar R50: 50: Reserve
    """
    BATTERY = "Battery"
    COMPRESSOR = "Compressor"
    DEVICE_INFORMATION = "DeviceInformation"
    DHWCIRCUIT = "DHWCircuit"
    DHWSTORAGE = "DHWStorage"
    DISHWASHER = "Dishwasher"
    DRYER = "Dryer"
    ELECTRICAL_IMMERSION_HEATER = "ElectricalImmersionHeater"
    FAN = "Fan"
    GAS_HEATING_APPLIANCE = "GasHeatingAppliance"
    METERING = "Metering"
    HEATING_BUFFER_STORAGE = "HeatingBufferStorage"
    HEATING_CIRCUIT = "HeatingCircuit"
    HEATING_OBJECT = "HeatingObject"
    HEATING_ZONE = "HeatingZone"
    HEAT_PUMP_CONTROL = "HeatPumpControl"
    HEAT_SINK_CIRCUIT = "HeatSinkCircuit"
    HEAT_SOURCE_CIRCUIT = "HeatSourceCircuit"
    HEAT_SOURCE_UNIT = "HeatSourceUnit"
    HVACCONTROLLER = "HVACController"
    HVACROOM = "HVACRoom"
    INSTANT_DHWHEATER = "InstantDHWHeater"
    INVERTER = "Inverter"
    OIL_HEATING_APPLIANCE = "OilHeatingAppliance"
    PUMP = "Pump"
    REFRIGERANT_CIRCUIT = "RefrigerantCircuit"
    SMART_ENERGY_APPLIANCE = "SmartEnergyAppliance"
    SOLAR_DHWSTORAGE = "SolarDHWStorage"
    SOLAR_THERMAL_CIRCUIT = "SolarThermalCircuit"
    SUB_METER_ELECTRICITY = "SubMeterElectricity"
    TEMPERATURE_SENSOR = "TemperatureSensor"
    WASHER = "Washer"
    BATTERY_SYSTEM = "BatterySystem"
    ELECTRICITY_GENERATION_SYSTEM = "ElectricityGenerationSystem"
    ELECTRICITY_STORAGE_SYSTEM = "ElectricityStorageSystem"
    SGCP = "SGCP"
    HOUSEHOLD = "Household"
    PVSYSTEM = "PVSystem"
    EV = "EV"
    EVSE = "EVSE"
    CHARGING_OUTLET = "ChargingOutlet"
    CEM = "CEM"
    ACTUATOR = "Actuator"
    SENSOR = "Sensor"
    CONTROLLER = "Controller"
    ENV_CONDITION = "EnvCondition"
    FLEX_BUILDING_CAMPUS = "FlexBuildingCampus"
    DYNAMIC_TARIFF = "DynamicTariff"
    R50 = "R50"



from dataclasses import dataclass, field
from typing import List, Optional
from sgrspecification.generic.data_point import DataPointBase
from sgrspecification.generic.functional_profile import FunctionalProfileBase
from sgrspecification.product.modbus_types import (
    ModbusAttributes,
    ModbusDataPointConfiguration,
    ModbusInterfaceDescription,
    TimeSyncBlockNotification,
)

__NAMESPACE__ = "http://www.smartgridready.com/ns/V0/"


@dataclass
class ModbusDataPoint(DataPointBase):
    """
    :ivar modbus_data_point_configuration:
    :ivar block_cache_identification: Refers to a
        timeSyncBlockNotification
    :ivar modbus_attributes:
    """
    modbus_data_point_configuration: Optional[ModbusDataPointConfiguration] = field(
        default=None,
        metadata={
            "name": "modbusDataPointConfiguration",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    block_cache_identification: Optional[str] = field(
        default=None,
        metadata={
            "name": "blockCacheIdentification",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    modbus_attributes: Optional[ModbusAttributes] = field(
        default=None,
        metadata={
            "name": "modbusAttributes",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class ModbusDataPointList:
    """
    List of data points.
    """
    data_point_list_element: List[ModbusDataPoint] = field(
        default_factory=list,
        metadata={
            "name": "dataPointListElement",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class ModbusFunctionalProfile(FunctionalProfileBase):
    modbus_attributes: Optional[ModbusAttributes] = field(
        default=None,
        metadata={
            "name": "modbusAttributes",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    data_point_list: Optional[ModbusDataPointList] = field(
        default=None,
        metadata={
            "name": "dataPointList",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


@dataclass
class ModbusFunctionalProfileList:
    """
    List of functional profiles.
    """
    functional_profile_list_element: List[ModbusFunctionalProfile] = field(
        default_factory=list,
        metadata={
            "name": "functionalProfileListElement",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class ModbusInterface:
    """
    Container for a modbus device.
    """
    modbus_interface_description: Optional[ModbusInterfaceDescription] = field(
        default=None,
        metadata={
            "name": "modbusInterfaceDescription",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    modbus_attributes: Optional[ModbusAttributes] = field(
        default=None,
        metadata={
            "name": "modbusAttributes",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    functional_profile_list: Optional[ModbusFunctionalProfileList] = field(
        default=None,
        metadata={
            "name": "functionalProfileList",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    time_sync_block_notification: List[TimeSyncBlockNotification] = field(
        default_factory=list,
        metadata={
            "name": "timeSyncBlockNotification",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )



import configparser
import re
from enum import Enum
from typing import Callable

from xsdata.formats.dataclass.context import XmlContext
from xsdata.formats.dataclass.parsers import XmlParser

from sgrspecification.product import DeviceFrame
from sgr_library.modbus_interface import SgrModbusInterface
from sgr_library.restapi_client_async import SgrRestInterface
from sgr_library.modbusRTU_interface_async import SgrModbusRtuInterface


class SGrConfiguration(Enum):
    UNKNOWN = 1
    STRING = 2
    FILE = 3


class SGrDeviceProtocol(Enum):
    MODBUS_RTU = 0
    MODBUS_TPC = 1
    RESTAPI = 2
    GENERIC = 3
    CONTACT = 4


def string_loader(xml_str: str) -> DeviceFrame:
    parser = XmlParser(context=XmlContext())
    return parser.from_string(xml_str, DeviceFrame)


def file_loader(path: str) -> DeviceFrame:
    parser = XmlParser(context=XmlContext())
    return parser.parse(path, DeviceFrame)


loaders: dict[SGrConfiguration, Callable[[str], DeviceFrame]] = {
    SGrConfiguration.STRING: string_loader,
    SGrConfiguration.FILE: file_loader,
}

device_builders: dict[SGrDeviceProtocol, Callable[
    [DeviceFrame, configparser.ConfigParser], SgrRestInterface | SgrModbusInterface | SgrModbusRtuInterface]
] = {
    SGrDeviceProtocol.MODBUS_TPC: lambda frame, _: SgrModbusInterface(frame),
    SGrDeviceProtocol.MODBUS_RTU: lambda frame, _: SgrModbusRtuInterface(frame),
    SGrDeviceProtocol.RESTAPI: lambda frame, config: SgrRestInterface(frame, config),
    # SGrDeviceProtocol.GENERIC: lambda frame, config: SgrModbusInterface(frame),
    # SGrDeviceProtocol.CONTACT: lambda frame, config: SgrModbusInterface(frame)
}


def resolve_protocol(frame: DeviceFrame) -> SGrDeviceProtocol:
    if frame.interface_list.rest_api_interface:
        return SGrDeviceProtocol.RESTAPI
    elif frame.interface_list.modbus_interface is not None and frame.interface_list.modbus_interface.modbus_interface_description.modbus_rtu:
        return SGrDeviceProtocol.MODBUS_RTU
    elif frame.interface_list.modbus_interface is not None and frame.interface_list.modbus_interface.modbus_interface_description.modbus_tcp:
        return SGrDeviceProtocol.MODBUS_TPC
    elif frame.interface_list.contact_interface:
        return SGrDeviceProtocol.CONTACT
    else:
        return SGrDeviceProtocol.GENERIC


class GenericSGrDeviceBuilder:

    def __init__(self):
        self._value: str | None = None
        self._config_value: str | dict | None = None
        self._type: SGrConfiguration = SGrConfiguration.UNKNOWN
        self._config_type: SGrConfiguration = SGrConfiguration.UNKNOWN

    def get_spec_content(self) -> str:
        if self._type == SGrConfiguration.FILE:
            try:
                input_file = open(self._value)
                return input_file.read()
            except:
                raise Exception("Invalid spec file path")
        elif self._type == SGrConfiguration.STRING:
            return self._value
        return ""

    def xml_file_path(self, file_path: str):
        self._value = file_path
        self._type = SGrConfiguration.FILE
        return self

    def xml_string(self, xml: str):
        self._value = xml
        self._type = SGrConfiguration.STRING
        return self

    def config_file_path(self, file_path: str):
        self._config_type = SGrConfiguration.FILE
        self._config_value = file_path
        return self

    def config(self, config: dict):
        self._config_type = SGrConfiguration.STRING
        self._config_value = config
        return self

    def replace_variables(self) -> tuple[str, configparser.ConfigParser]:
        config = configparser.ConfigParser()
        if self._config_type is SGrConfiguration.FILE:
            config.read(self._config_value)
        elif self._config_type is SGrConfiguration.STRING:
            config.read_dict(self._config_value)
        spec = self.get_spec_content()
        for section_name, section in config.items():
            for param_name in section:
                pattern = re.compile(r'{{' + param_name + r'}}')
                spec = pattern.sub(config.get(section_name, param_name), spec)

        return spec, config

    def build(self) -> SgrRestInterface | SgrModbusInterface | SgrModbusRtuInterface:
        if self._type not in loaders:
            raise Exception(f'unsupported loader configuration, {self._type}')

        spec, config = self.replace_variables()
        xml = loaders[SGrConfiguration.STRING](spec)
        protocol = resolve_protocol(xml)
        return device_builders[protocol](xml, config)



from enum import Enum


class DataTypes(Enum):
    INT = 'INT',
    FLOAT = 'FLOAT',
    STRING = 'STRING',
    ENUM = 'ENUM',
    BOOLEAN = 'BOOLEAN',
    BITMAP = 'BITMAP',
    DATE_TIME = 'DATE_TIME',



import asyncio

from pymodbus.client import AsyncModbusSerialClient
from pymodbus.constants import Endian

from sgr_library.payload_decoder import PayloadDecoder, PayloadBuilder, RoundingScheme


# In this case establishes a connection with the localhost server that is running the simulation.

class SGrModbusRTUClient:

    def __init__(self, port: str, parity: str, baudrate: int, client=None):
        """
        Creates client
        :param ip: The host to connect to (default 127.0.0.1)
        :param port: The modbus port to connect to (default 502)
        """
        self._port = port
        self._semaphore = asyncio.Semaphore(1)
        if client is not None:
            self.client = client
        else:
            self.client = AsyncModbusSerialClient(
                method="rtu",
                port=port,
                parity=parity,
                baudrate=baudrate,
            )  # changed source: https://stackoverflow.com/questions/58773476/why-do-i-get-pymodbus-modbusioexception-on-20-of-attempts

            # connected = self.client.connect() #TODO a wrapper that opens and closes connection when function is excecuted?
            # if connected:
            #     print("Connected to ModbusRTU ond Port: " + port)

    async def connect(self):
        await self.client.connect()
        print("Connected to ModbusRTU ond Port: " + self._port)

    # not changed
    async def value_decoder(self, addr: int, size: int, data_type: str, register_type: str, slave_id: int,
                            order: Endian) -> float:
        """
        Reads register and decodes the value.
        :param addr: The address to read from and decode
        :param size: The number of registers to read
        :param data_type: The modbus type to decode
        :returns: Decoded float
        """
        async with self._semaphore:
            if register_type == "HoldRegister":  # Todo im Modbus_client ist "HoldingRegister" angeben, ist das falsch?
                reg = await self.client.read_holding_registers(addr, count=size, slave=slave_id)
                # print(reg.registers)
            else:
                reg = await self.client.read_input_registers(addr, count=size, slave=slave_id)
            decoder = PayloadDecoder.fromRegisters(reg.registers, byteorder=order, wordorder=order)

            if not reg.isError():
                return decoder.decode(data_type, 0)

    # not changed
    async def value_encoder(self, addr: int, value: float, data_type: str, slave_id: int, order: Endian):
        """
        Encodes value to be set on the register address
        :param addr: The address to read from and decode
        :param value: The value to be written on the register
        :param data_type: The modbus type to decode
        """
        async with self._semaphore:
            builder = PayloadBuilder(byteorder=order, wordorder=order)
            builder.encode(value, data_type, rounding=RoundingScheme.floor)
            await self.client.write_registers(address=addr, values=builder.to_registers(), unit=slave_id)


if __name__ == "__main__":
    async def run():
        MyClient = SGrModbusRTUClient("COM5", "E", 19200)

        connected = await MyClient.client.connect()

        a = await MyClient.value_decoder(0x5B14, 2, "int32", "HoldRegister", slave_id=1, order=Endian.BIG)
        print(f"Der Wert ist: {a}")
        await MyClient.client.close()


    asyncio.run(run())



from sgrspecification.product import ConfigurationListElement, ConfigurationList


def build_configurations_parameters(params: ConfigurationList):
    if params is None:
        return []
    return [ConfigurationParameter(x) for x in params.configuration_list_element]


class ConfigurationParameter:

    def __init__(self, parameter: ConfigurationListElement):
        translation = parameter.configuration_description
        self.label = translation[0].label
        self.match = parameter.name
        self.type = parameter.data_type
        self.description = translation[0].text_element

    def __str__(self):
        return f'{self.label} - {self.match}'

    def __repr__(self):
        return f'{self.label} - {self.match}'



"""
SGr Payload Builder
------------------------

An adapted payload decoder and payload builder to use with the pymodbus library.
"""
from pymodbus.payload import BinaryPayloadDecoder, BinaryPayloadBuilder
from enum import Enum
from math import ceil, floor


class RoundingScheme(Enum):
    floor = 'floor'
    ceil = 'ceil'
    near = 'Near'


def round_to_int(value: float, scheme: RoundingScheme) -> int:
    if scheme == RoundingScheme.floor:
        return floor(value)
    elif scheme == RoundingScheme.ceil:
        return ceil(value)
    elif scheme == RoundingScheme.near:
        return round(value)
    else:
        print(
            'tried rounding with a invalid scheme (%s) using floor instead', scheme
        )
        return int(floor(value))


class PayloadDecoder(BinaryPayloadDecoder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def decode(self, modbus_type: str, byte_count: int):
        """
        :param modbus_type: 'int8', 'int8_u', 'int16', 'int16_u', 'int32', 'int32_u', 'int64', 'int64_u', 'float16', 'float32', 'float64', 'string'
        """
        # TODO boolean, enum, date_time
        if modbus_type == 'int8':
            return self.decode_8bit_int()
        elif modbus_type == 'int8_u':
            return self.decode_8bit_uint()
        elif modbus_type == 'int16':
            return self.decode_16bit_int()
        elif modbus_type == 'int16_u':
            return self.decode_16bit_uint()
        elif modbus_type == 'int32':
            return self.decode_32bit_int()
        elif modbus_type == 'int32_u':
            return self.decode_32bit_uint()
        elif modbus_type == 'int64':
            return self.decode_64bit_int()
        elif modbus_type == 'int64_u':
            return self.decode_64bit_uint()
        elif modbus_type == 'float16':
            return self.decode_16bit_float()
        elif modbus_type == 'float32':
            return self.decode_32bit_float()
        elif modbus_type == 'float64':
            return self.decode_64bit_float()
        elif modbus_type == 'string':
            return self.decode_string(byte_count)
        else:
            raise ValueError('Unknown modbus type "%s"', modbus_type)


class PayloadBuilder(BinaryPayloadBuilder):
    def __init__(self, *args, **kwarg):
        super().__init__(*args, **kwarg)

    def encode(self, value: float, modbus_type: str, rounding: RoundingScheme):
        """
        :param modbus_type: 'int8', 'int8_u', 'int16', 'int16_u', 'int32', 'int32_u', 'int64', 'int64_u', 'float16', 'float32', 'float64', 'string'
        """
        # TODO boolean, int8, int8_u, enum, date_time
        if modbus_type == 'int8':
            self.add_8bit_int(round_to_int(value, rounding))
        elif modbus_type == 'int8_u':
            self.add_8bit_uint(round_to_int(value, rounding))
        elif modbus_type == 'int16':
            self.add_16bit_int(round_to_int(value, rounding))
        elif modbus_type == 'int16_u':
            self.add_16bit_uint(round_to_int(value, rounding))
        elif modbus_type == 'int32':
            self.add_32bit_int(round_to_int(value, rounding))
        elif modbus_type == 'int32_u':
            self.add_32bit_uint(round_to_int(value, rounding))
        elif modbus_type == 'int64':
            self.add_64bit_int(round_to_int(value, rounding))
        elif modbus_type == 'int64_u':
            self.add_64bit_uint(round_to_int(value, rounding))
        elif modbus_type == 'float16':
            self.add_16bit_float(round_to_int(value, rounding))
        elif modbus_type == 'float32':
            self.add_32bit_float(round_to_int(value, rounding))
        elif modbus_type == 'float64':
            self.add_64bit_float(round_to_int(value, rounding))
        elif modbus_type == 'string':
            self.add_string(round_to_int(value, rounding))
        else:
            print('Unknown modbus type "%s"', modbus_type)



from sgr_library.payload_decoder import PayloadDecoder, PayloadBuilder
from pymodbus.constants import Endian
from pymodbus.client import ModbusTcpClient, ModbusSerialClient
from sgr_library.modbus_interface import SgrModbusInterface

# In this case establishes a connection with the localhost server that is running the simulation.

class SGrModbusRTUClient:

    def __init__(self, port: str, parity: str, baudrate: int):
        """
        Creates client
        :param ip: The host to connect to (default 127.0.0.1)
        :param port: The modbus port to connect to (default 502)
        """
        #EXAMPLE: client = ModbusTcpClient('127.0.0.1', 5002)
        self.client = ModbusSerialClient(method="rtu", port=port, parity=parity, baudrate=baudrate, timeout=1) #changed source: https://stackoverflow.com/questions/58773476/why-do-i-get-pymodbus-modbusioexception-on-20-of-attempts
        connected = self.client.connect() #TODO a wrapper that opens and closes connection when function is excecuted?
        if connected:
            print("Connected to ModbusRTU ond Port: " + port)

    #not changed
    def value_decoder(self, addr: int, size: int, data_type: str, register_type: str, slave_id: int, order: Endian) -> float:
        """
        Reads register and decodes the value.
        :param addr: The address to read from and decode
        :param size: The number of registers to read
        :param data_type: The modbus type to decode
        :returns: Decoded float
        """
        if register_type == "HoldRegister": #Todo im Modbus_client ist "HoldingRegister" angeben, ist das falsch?
            reg = self.client.read_holding_registers(addr, count=size, unit=slave_id)
            #print(reg.registers)
        else:
            reg = self.client.read_input_registers(addr, count=size, unit=slave_id)
        decoder = PayloadDecoder.fromRegisters(reg.registers, byteorder=order, wordorder=order)
        
        if not reg.isError():
            return decoder.decode(data_type, 0)
    #not changed
    def value_encoder(self, addr: int, value: float, data_type: str, slave_id: int, order: Endian):
        """
        Encodes value to be set on the register address
        :param addr: The address to read from and decode
        :param value: The value to be written on the register
        :param data_type: The modbus type to decode
        """
        builder = PayloadBuilder(byteorder=order, wordorder=order)
        builder.encode(value, data_type, rounding="floor")
        self.client.write_registers(address=addr, values=builder.to_registers(), unit=slave_id)


if __name__ == "__main__":

    #Build Connection
    myRTUModbus = SGrModbusRTUClient("COM7", "E", 19200)

    #int16_u --> 16bit pro register, signedvalue
    Power = myRTUModbus.value_decoder(0x5B14,2,"int32","HoldingRegister",1,Endian.Big)

    print(str(Power*0.01) + "W")

    myRTUModbus.client.close()
    print("finished")



from setuptools import setup


def read_requirements() -> list[str]:
    with open('./requirements.txt', 'r') as file:
        return file.read().splitlines()


setup(
    name='SGrPythontks4r',
    use_scm_version={'local_scheme': 'no-local-version'},
    setup_requires=['setuptools_scm'],
    packages=['sgr_library', 'sgr_library.generated', 'sgr_library.generated.generic',
              'sgr_library.generated.product', 'sgr_library.generated.communicator',
              'sgr_library.generated.functional_profile', 'sgr_library.api', 'sgr_library.validators',
              'sgr_library.converters'],
    url='https://github.com/robin-schoch/SGrPython',
    license='',
    author='robin',
    author_email='',
    description='',
    install_requires=read_requirements(),
    long_description='',
    long_description_content_type='text/plain',
)



from sgr_library.payload_decoder import PayloadDecoder, PayloadBuilder, RoundingScheme
from pymodbus.constants import Endian
from pymodbus.client import AsyncModbusTcpClient
from typing import Optional, Tuple, Dict, Any, Iterable
from sgr_library.exceptions import RegisterError
import asyncio
import logging

# In this case establishes a connection with the localhost server that is running the simulation.
# TODO make this inherit from the ModbusTcpClient
# TODO maybe make everything in one module... Modulating this is redundant in my taste.

class SGrModbusClient:

    def __init__(self, ip: str, port: int):
        """
        Creates client
        :param ip: The host to connect to (default 127.0.0.1)
        :param port: The modbus port to connect to (default 502)
        """
        #EXAMPLE: client = ModbusTcpClient('127.0.0.1', 5002)
        self.client = AsyncModbusTcpClient(
        host=ip,
        port=port,
        timeout=1,
        retries=0,
        reconnect_delay=5000,
        close_comm_on_error=False,)
        #self.client.connect() #TODO a wrapper that opens and closes connection when function is excecuted?

    async def value_decoder(self, addr: int, size: int, data_type: str, register_type: str, slave_id: int, order: Endian) -> Optional[float]:
        """
        Reads register and decodes the value.
        :param addr: The address to read from and decode
        :param size: The number of registers to read
        :param data_type: The modbus type to decode
        :param register_type: The type of the register
        :param slave_id: The ID of the slave
        :param order: The endian order
        :returns: Decoded float
        """
        try:
            if register_type == "HoldRegister":
                reg = await self.client.read_holding_registers(addr, size, slave=slave_id)
            elif register_type == "InputRegister":
                reg = await self.client.read_input_registers(addr, count=size, slave=slave_id)
            else:
                raise ValueError(f"Invalid register type: {register_type}")

            if reg.isError():
                raise RegisterError(f"Error reading register: {reg}")


            decoder = PayloadDecoder.fromRegisters(reg.registers, byteorder=order, wordorder=order)
            return decoder.decode(data_type, 0)
            # TODO: Add code to decode and return the float value based on the data type
        except asyncio.TimeoutError:
            logging.exception(f"Timeout reading register at address {addr} with slave ID {slave_id}")
            return None
        except RegisterError as e:
            logging.exception(str(e))
            return None
        except Exception as e:
            logging.exception(f"An unexpected error occurred: {e}")
            return None

    # TODO Under construction
    async def mult_value_decoder(self, addr: int, size: int, data_type: str, register_type: str, slave_id: int, order: Endian) -> Optional[float]:
        """
        Reads register and decodes the value.
        :param addr: The address to read from and decode
        :param size: The number of registers to read
        :param data_type: The modbus type to decode
        :returns: Decoded float
        """
        if register_type == "HoldRegister":
            reg = await self.client.read_holding_registers(addr, size, slave=slave_id) #TODO add slave id?
        else:
            reg = await self.client.read_input_registers(addr, count=size, slave=slave_id)
        decoder = PayloadDecoder.fromRegisters(reg.registers, byteorder=order, wordorder=order)
        #print(decoder.decode('float32', 0))
        if not reg.isError():
            #await print(decoder.decode(data_type, 0))
            indexes = [size//3*0, size//3*1, size//3*2]
            l1 = decoder.decode(data_type, indexes[0])
            l2 = decoder.decode(data_type, indexes[1])
            l3 = decoder.decode(data_type, indexes[2])
            return l1, l2, l3


    async def value_encoder(self, addr: int, value: float, data_type: str, slave_id: int, order: Endian):
        """
        Encodes value to be set on the register address
        :param addr: The address to read from and decode
        :param value: The value to be written on the register
        :param data_type: The modbus type to decode
        :param slave_id: The ID of the slave
        :param order: The endian order
        """
        try:
            builder = PayloadBuilder(byteorder=order, wordorder=order)
            builder.encode(value, data_type, rounding=RoundingScheme.floor)
            self.client.write_registers(address=addr, values=builder.to_registers(), unit=slave_id)
        except asyncio.TimeoutError:
            logging.exception(f"Timeout writing value to register at address {addr} with slave ID {slave_id}")
        except ValueError as e:
            logging.exception(f"Value error occurred: {e}")
        except Exception as e:
            logging.exception(f"An unexpected error occurred while writing value to register: {e}")



# nothing here
from sgrspecification import (
    communicator,
    functional_profile,
    generic,
    product
)

__all__ = [
    "communicator",
    "functional_profile",
    "generic",
    "product"
]



from sgrspecification.functional_profile.functional_profile_frame import (
    FunctionalProfileDataPoint,
    FunctionalProfileDataPointList,
    FunctionalProfileFrame,
)

__all__ = [
    "FunctionalProfileDataPoint",
    "FunctionalProfileDataPointList",
    "FunctionalProfileFrame",
]




from sgr_library import SGrDevice
import asyncio


async def test_loop():

    config_file = 'config_wago'
    interface_file = 'abb_terra_01.xml'

    config_file = 'config_CLEMAPEnMon_ressource_default.ini'
    interface_file = 'SGr_04_mmmm_dddd_CLEMAPEnergyMonitorEIV0.2.1.xml'

    

    device = SGrDevice()
    device.update_xml_spec(interface_file).update_config(config_file).build()
    await device.connect()

    #device_data = await device.read_data()
    #print(device_data)

    fp = device.get_function_profile("PowerFactor") #PowerFactor #VoltageAC
    #fp_data = await fp.read()
    #print(fp_data)

    dp = fp.get_data_point("PowerFactorTOT") #PowerFactorTOT #VoltageACL1
    dp_data = await dp.read()
    print(dp_data)

#VoltageDC_IN_1
#VoltageDC
#SGr_04_0014_0000_WAGO_Testsystem_V1.0.xml

try:
    asyncio.run(test_loop())
except KeyboardInterrupt:

    # Here we have to close all the sessions...
    # We have to think if we want to open a connection and close it for
    # every getval, or we just leave the user do this.
    print("done")


from typing import Callable

from sgr_library.api import DataPointConverter
from sgr_library.converters.converter import HoursConverter, MinutesConverter, NoneConverter, SecondsConverter, TemperatureConverter, VoltDataPointConverter, UnsupportedConverter, \
    ReactiveEnergyConverter, PowerOverTimeConverter, CurrentConverter, PowerConverter, PercentConverter
from sgrspecification.generic import Units

converter_lookup: dict[Units, Callable[[Units], DataPointConverter]] = {
    Units.WATTS: lambda x: PowerConverter(x),
    Units.KILOWATTS: lambda x: PowerConverter(x),
    Units.VOLTS: lambda x: VoltDataPointConverter(x),
    Units.AMPERES: lambda x: CurrentConverter(x),
    Units.KILOVOLT_AMPERES_REACTIVE: lambda x: ReactiveEnergyConverter(x),
    Units.KILOVOLT_AMPERES: lambda x: ReactiveEnergyConverter(x),
    Units.KILOWATT_HOURS: lambda x: PowerOverTimeConverter(x),
    Units.KILOVOLT_AMPERES_REACTIVE_HOURS: lambda x: ReactiveEnergyConverter(x),
    Units.VOLT_AMPERES_REACTIVE: lambda x: ReactiveEnergyConverter(x),
    Units.HERTZ: lambda x : ReactiveEnergyConverter(x),
    Units.VOLT_AMPERES: lambda x: ReactiveEnergyConverter(x),

    Units.DEGREES_CELSIUS: lambda x: TemperatureConverter(x),

    Units.HOURS: lambda x: HoursConverter(x),
    Units.MINUTES: lambda x: MinutesConverter(x),
    Units.SECONDS: lambda x: SecondsConverter(x),

    Units.PERCENT: lambda x: PercentConverter(x),

    Units.NONE: lambda x: NoneConverter(x),

}


def build_converter(unit: Units) -> DataPointConverter:
    return converter_lookup.get(unit, lambda x: UnsupportedConverter(x))(unit)



from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional
from sgrspecification.generic.base_type_level_of_operation_type import LevelOfOperation
from sgrspecification.generic.base_types import (
    AlternativeNames,
    GenericAttributeListFunctionalProfile,
    LegibleDescription,
    ReleaseNotes,
)
from sgrspecification.product.modbus_types import ModbusInterfaceSelection
from sgrspecification.product.rest_api_types import RestApiInterfaceSelection

__NAMESPACE__ = "http://www.smartgridready.com/ns/V0/"


class CommunicatorCategory(Enum):
    """
    Communicator types, a growing list of supported Communicators.
    """
    SMART_HOME = "SmartHome"
    LIGHT_CONTROL = "LightControl"
    POWER_BALANCER = "PowerBalancer"
    ACCOUNTING_SYSTEM = "AccountingSystem"


@dataclass
class CommunicatorTransportService:
    modbus: Optional[ModbusInterfaceSelection] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    rest: Optional[RestApiInterfaceSelection] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class TransportServices:
    """
    A list of supported transport services.
    """
    transport_service: List[CommunicatorTransportService] = field(
        default_factory=list,
        metadata={
            "name": "transportService",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class CommunicatorInformation:
    """
    :ivar alternative_names:
    :ivar legible_description: Published and printable information
        related to this communicator
    :ivar supported_transport_services:
    :ivar communicator_category:
    :ivar version_number:
    :ivar brand_name: branding information
    :ivar manufacturer_specification_identification: specification
        identifier
    :ivar manufacturer_label: the label of the device
    :ivar general_remarks: author of this sheet may add remarks / non
        disclaimer statements
    :ivar level_of_operation: defines the SGr Label Level 1...6 of the
        highest level functional profile supported by the communicator
    :ivar is_local_control: Value "false" means "is cloud control
        device", indicating that this service is based on cloud. "True"
        indicates that services are provided within the range of the
        local area.
    """
    alternative_names: Optional[AlternativeNames] = field(
        default=None,
        metadata={
            "name": "alternativeNames",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    legible_description: List[LegibleDescription] = field(
        default_factory=list,
        metadata={
            "name": "legibleDescription",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "max_occurs": 4,
        }
    )
    supported_transport_services: Optional[TransportServices] = field(
        default=None,
        metadata={
            "name": "supportedTransportServices",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    communicator_category: Optional[CommunicatorCategory] = field(
        default=None,
        metadata={
            "name": "communicatorCategory",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    version_number: Optional[str] = field(
        default=None,
        metadata={
            "name": "versionNumber",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    brand_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "brandName",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    manufacturer_specification_identification: Optional[str] = field(
        default=None,
        metadata={
            "name": "manufacturerSpecificationIdentification",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    manufacturer_label: Optional[str] = field(
        default=None,
        metadata={
            "name": "manufacturerLabel",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    general_remarks: Optional[str] = field(
        default=None,
        metadata={
            "name": "generalRemarks",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    level_of_operation: Optional[LevelOfOperation] = field(
        default=None,
        metadata={
            "name": "levelOfOperation",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    is_local_control: Optional[bool] = field(
        default=None,
        metadata={
            "name": "isLocalControl",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


@dataclass
class CommunicatorBase:
    """
    Base type for device.

    :ivar communicator_name: Device Name in the context of the
        manufacturer
    :ivar manufacturer_name: Name of the Manufacturer or OEM
    :ivar release_notes:
    :ivar specification_owner_identification: the identifier as
        enumeration indicates that the manufacturer is related with the
        organisation and that this declaration is generated by himself
    :ivar communicator_information:
    :ivar generic_attribute_list:
    """
    communicator_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "communicatorName",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    manufacturer_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "manufacturerName",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    release_notes: Optional[ReleaseNotes] = field(
        default=None,
        metadata={
            "name": "releaseNotes",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    specification_owner_identification: Optional[str] = field(
        default=None,
        metadata={
            "name": "specificationOwnerIdentification",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    communicator_information: Optional[CommunicatorInformation] = field(
        default=None,
        metadata={
            "name": "communicatorInformation",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    generic_attribute_list: Optional[GenericAttributeListFunctionalProfile] = field(
        default=None,
        metadata={
            "name": "genericAttributeList",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
